#include "Wizard.h"
#include "Beast.h"
#include "Utils.h"
#include <math.h>
#include <stdio.h>

Wizard::Wizard(Detector* detector) : Object(detector, "Models/wizard-rest.obj", "Textures/wizzow.tga") {
    model->setHeadingCorrection(90);
    model->setScale(0.5);
    Frame* frame0 = new Frame("Models/wizard-rest.obj", 0.2);
    Frame* frame1 = new Frame("Models/wizard-walk1.obj", 0.2);
    Frame* frame2 = new Frame("Models/wizard-walk2.obj", 0.2);
    Frame* frame3 = new Frame("Models/wizard-cast.obj", 0.1);
    Frame* frame4 = new Frame("Models/wizard-dead.obj", 1);
    Animation* walkAnimation = new Animation(model, true);
    Animation* idleAnimation = new Animation(model, false);
    Animation* castAnimation = new Animation(model, false);
    Animation* deathAnimation = new Animation(model, false);
    walkAnimation->addFrame(frame1);
    walkAnimation->addFrame(frame0);
    walkAnimation->addFrame(frame2);
    walkAnimation->addFrame(frame0);
    idleAnimation->addFrame(frame0);
    castAnimation->addFrame(frame3);
    deathAnimation->addFrame(frame4);
    model->addAnimation(WIZARD_WALK, walkAnimation);
    model->addAnimation(WIZARD_REST, idleAnimation);
    model->addAnimation(WIZARD_CAST, castAnimation);
    model->addAnimation(WIZARD_DEAD, deathAnimation);
    model->setCurrentAnimation(WIZARD_REST);
    
    health = 10;
    maxHealth = 10;
    mana = 30;
    maxMana = 99;
    gold = 15;
    regenerateHTime = 0;
    regenerateMTime = 0;
    animating = false;
    fireballLevel = 1;
    sprayLevel = 1;
    
    spray = NULL;
    sprayTime = 0;
    alive = true;
}

void Wizard::moveForward(float deltaTime) {
    Object::moveForward(3 * deltaTime);
    if (model->isAnimating() && (model->getCurrentAnimation() != WIZARD_WALK)) {
        model->setCurrentAnimation(WIZARD_WALK);
    } else if (!model->isAnimating()) {
        model->setCurrentAnimation(WIZARD_WALK);
        model->startAnimation();
    }
}

void Wizard::moveBackward(float deltaTime) {
    Object::moveBackward(3 * deltaTime);
    if (model->isAnimating() && (model->getCurrentAnimation() != WIZARD_WALK)) {
        model->setCurrentAnimation(WIZARD_WALK);
    } else if (!model->isAnimating()) {
        model->setCurrentAnimation(WIZARD_WALK);
        model->startAnimation();
    }
}

void Wizard::castAnimation() {
    if (model->isAnimating() && (model->getCurrentAnimation() != WIZARD_CAST)) {
        model->setCurrentAnimation(WIZARD_CAST);
    } else if (!model->isAnimating()) {
        if (model->getCurrentAnimation() != WIZARD_CAST) {
            model->setCurrentAnimation(WIZARD_CAST);
            model->startAnimation();
        }
    }
}

void Wizard::stop() {
    model->setCurrentAnimation(WIZARD_REST);
    if (!model->isAnimating()) {
        model->startAnimation();   
    }
    
    if (spray != NULL) {
        delete spray;   
        spray = NULL; 
    }
    
    animating = false;
}

void Wizard::changeHealth(int amount) {
    if (health < maxHealth - amount) {
        health+= amount;
    } else {
        health = maxHealth;
    }
    if (health <= 0) {
        health = 0;
        die();
    }
    
    if (amount < 0) {
        textColor[0] = 1;
        textColor[1] = 0;
        textColor[2] = 0;    
    } else {
        textColor[0] = 0;
        textColor[1] = 1;
        textColor[2] = 0;  
    }
    
    char msg[10];
    sprintf(msg, "%d", amount);
    showText(msg, 0.5);
}

void Wizard::die() {
    model->setCurrentAnimation(WIZARD_DEAD);
    animating = true;
    alive = false;
}

bool Wizard::isAlive() {
    return alive;    
}

void Wizard::pickup(Gold* goldObj) {
    if (gold < 99) {
        gold += 1;
    }
    delete goldObj;
}

int Wizard::getHealth() {
    return health;
}

int Wizard::getMana() {
    return mana;
}

int Wizard::getGold() {
    return gold;
}

int Wizard::getMaxHealth() {
    return maxHealth;
}

void Wizard::upgradeSpray() {
    if (gold >= 5) {
        if (sprayLevel < 10)
            gold -= 5;
            sprayLevel += 1;
    }
}

void Wizard::upgradeFireball() {
    if (gold >= 5) {
        if (fireballLevel < 10) {
            gold -= 5;
            fireballLevel += 1;
        }
    }
}

void Wizard::upgradeHealth() {
    if (gold >= 1) {
        gold--;
        maxHealth += 1;
    }
}

void Wizard::idle(float deltaTime) {
    Object::idle(deltaTime);
    if (alive) {
        regenerateHTime += deltaTime;
        regenerateMTime += deltaTime;
        if (regenerateHTime > 10) {
            if (health < maxHealth) {
                changeHealth(1);
            }
            regenerateHTime = 0;
        }
        if (regenerateMTime > 1) {
            if (mana < maxMana) {
                mana += 1;
            }
            regenerateMTime = 0;
        }
        
        if (spray != NULL) {
            if (mana > 5) {
                sprayTime += deltaTime;
                            
                if (sprayTime > 0.5) { //decrease mana
                    mana -= 5;   
                    sprayTime = 0; 
                } else {
                    spray->idle(deltaTime);
                }
            } else {
                stop();
            }
        }
    }
    
    std::list<Fireball*>::iterator i;
    for (i = fireballs.begin(); i != fireballs.end(); i++) {
        Fireball* obj = *i;
        if (obj->isMarked()) {
            i--;
            delete obj;
            if (fireballs.size() == 0) {
                stop();
            }
        } else {
            obj->idle(deltaTime);
        }
    }
}

void Wizard::castFireball(Enemy* beast) {
    if (mana < 10) {
        return;
    }
    mana -= 10;
    Fireball* fireball = new Fireball(detector, fireballLevel, this);
    fireballs.push_back(fireball);
    
    float x = beast->getPositionX();
    float y = beast->getPositionY();
    float z = beast->getPositionZ();
    float ballOrientation;
    orientation = ballOrientation = 90 + atan2f((x-positionX),(z-positionZ))*180.0/M_PI;
    
    fireball->setPosition(positionX,1,positionZ,ballOrientation);
    fireball->moveForward(3);
    animating = true;
    castAnimation();
}

void Wizard::castSpray() {
    if (spray == NULL) {
        spray = new Spray(detector, sprayLevel, this);
        castAnimation();    
    }
}

void Wizard::draw() {
    glColor3f(1,1,1);
    Object::draw();
    std::list<Fireball*>::iterator i;
    for (i = fireballs.begin(); i != fireballs.end(); i++) {
        Fireball* obj = *i;
        obj->draw();
    }    
    
    if (alive) {
        if (spray != NULL) {
            spray->draw();   
        }
    }
}

void Wizard::removeFireball(Fireball* fireball) {
    fireballs.remove(fireball);
}

int Wizard::getSprayLevel() {
    return sprayLevel;
}

int Wizard::getFireballLevel() {
    return fireballLevel;
}

bool Wizard::isAnimating() {
    return animating;   
}

void Wizard::setAnimating(bool anim) {
    animating = anim;   
}

void Wizard::getBoundingVolume(float *minX,float *maxX,
                           float *minY,float *maxY,
                           float *minZ,float *maxZ) {
    Object::getBoundingVolume(minX,maxX,minY,maxY,minZ,maxZ);
    *minX += 0.1;
    *maxX -= 0.1;
    *minY += 0.1;
    *maxY -= 0.1;
    *minZ += 0.1;
    *maxZ -= 0.1;
}
