#include <GL/glut.h>
#include <stdlib.h>
#include <iostream>
#include "Camera.h"
#include "CameraState.h"
#include "Utils.h"
#include "Skybox.h"
#include "RainParticleSystem.h"
#include "Wizard.h"
#include "Level.h"
#include "Hud.h"
#include <time.h>

/*******************************************************************************
 * GLUT Callbacks
 ******************************************************************************/
void KeyboardCallback(unsigned char c, int x, int y);
void KeyboardUpCallback(unsigned char c, int x, int y);
void SpecialKeyCallback(int key, int x, int y);
void SpecialKeyUpCallback(int key, int x, int y);
void MouseMotionCallback(int x, int y);
void MouseCallback(int button, int state, int x, int y);
void DisplayCallback();
void IdleCallback();

/*******************************************************************************
 * Keyboard/Mouse globals
 ******************************************************************************/
 
//ASCII key press state information (true if pressed, false otherwise)
bool isPressed[256];     
//left mouse button pressed state
bool leftButtonDown = false;    
//right mouse button pressed state
bool rightButtonDown = false;
//middle mouse button presse state
bool middleButtonDown = false;
//up arrow key pressed state
bool upPressed = false;
//down arrow key pressed state
bool downPressed = false;
//left arrow key pressed state
bool leftPressed = false;
//right arrow key pressed state
bool rightPressed = false;
//where the last mouse click was
int pressX = 0;
int pressY = 0;
int currentX = 0;
int currentY = 0;
int prevX = 0;
int prevY = 0;

/*******************************************************************************
 * Elapsed Time
 ******************************************************************************/
 
//time that the idle call was last called
int timeLastIdle = 0;
//frames per second (updated once per second)
int fps = 0;
//frame counter
int fpsCounter = 0;
//frames per second time counter (time passed since fps was last updated)
int fpsTimeCounter = 0;

/*******************************************************************************
 * Lighting stuff
 ******************************************************************************/
GLfloat light0Pos[4] = {-10,10,10,0};
GLfloat diffuse0[4] = {0.3,0.3,0.3,1};
GLfloat ambient0[4] = {0.3,0.3,0.3,1};
GLfloat diffuse1[4] = {1,1,1,1};
GLfloat ambient1[4] = {0,0,0,1};
GLfloat specular[4] = {0,0,0,0};

/*******************************************************************************
 * Picking stuff
 ******************************************************************************/
int pick(int x, int y);

/*******************************************************************************
 * Game globals
 ******************************************************************************/

Detector* detector;
Wizard* wizard;
Level* level;
Hud* hud;

Skybox* skyBox;
CameraState camState;
Camera cam(&camState);

RainParticleSystem* rain;

/*******************************************************************************
 ******************************************************************************/

void KeyboardCallback(unsigned char c, int x, int y) {
    //flag the currently pressed key as being pressed
    isPressed[c] = true;
}

void KeyboardUpCallback(unsigned char c, int x, int y) {
    //flag the currently pressed key as being not pressed
    isPressed[c] = false;
}

void SpecialKeyCallback(int key, int x, int y) {
    //check if an arrow key was pressed
    if (key == GLUT_KEY_UP) {
        upPressed = true;
    } else if (key == GLUT_KEY_DOWN) {
        downPressed = true;
    } else if (key == GLUT_KEY_LEFT) {
        leftPressed = true;
    } else if (key == GLUT_KEY_RIGHT) {
        rightPressed = true;
    }
}

void SpecialKeyUpCallback(int key, int x, int y) {
    //check if an arrow key was released
    if (key == GLUT_KEY_UP) {
        upPressed = false;
    } else if (key == GLUT_KEY_DOWN) {
        downPressed = false;
    } else if (key == GLUT_KEY_LEFT) {
        leftPressed = false;
    } else if (key == GLUT_KEY_RIGHT) {
        rightPressed = false;
    }
}

void IdleCallback() {    
    //Calculate the time that has elapsed since the last call to IdleCallback
    int currentTime = glutGet(GLUT_ELAPSED_TIME);
    int delta = currentTime - timeLastIdle;
    if (timeLastIdle == 0) { //This only happens the first time IdleCallback is called
         delta = 0;
    }
    timeLastIdle = currentTime;
    //scaled value of the time delta. in seconds.
    float deltaTime = delta / 1000.0;
    
    //Frames per second calculation
    fpsCounter++;
    fpsTimeCounter += delta;
    if (fpsTimeCounter >= 1000) {
        fps = fpsCounter;
        fpsCounter = 0;
        fpsTimeCounter = 0;
        std::cout << "FPS: " << fps << std::endl;
    }
    
    if (isPressed['q'] || isPressed['Q']) {
        exit(0);    
    }    
    
    if (wizard->isAlive()) {
        if (isPressed[' ']) {
            wizard->castSpray();
        } else {
            if (!wizard->isAnimating()) {
                if (upPressed || isPressed['w']) {
                    wizard->moveForward(deltaTime);      
                } else if (downPressed || isPressed['s']) {
                    wizard->moveBackward(deltaTime);
                } else {
                    wizard->stop();   
                }
            }
        }
        
        if (leftPressed || isPressed['a']) {
            wizard->rotateCCW(deltaTime);
        }
        if (rightPressed || isPressed['d']) {
            wizard->rotateCW(deltaTime);
        }
        
        if (leftButtonDown) {
            leftButtonDown = false;
            int picked = pick(pressX, pressY);   
            if (picked != -1) {
                Enemy* beast = detector->findEnemyByNum(picked);
                wizard->castFireball(beast);
            }
        }
        
        level->idle(deltaTime);
        detector->idle(deltaTime);        
    } else {
        camState.around_target_rot_y += 45 * deltaTime;    
    }
    rain->idle(deltaTime);
    wizard->idle(deltaTime);

    camState.target_x = wizard->getPositionX();
    camState.target_y = wizard->getPositionY();
    camState.target_z = wizard->getPositionZ(); 

    glutPostRedisplay();
}

void MouseMotionCallback(int x, int y) {
    prevX = currentX;
    prevY = currentY;
    currentX = x;
    currentY = y;
    
    if (x < prevX) {
        if (rightButtonDown) {
            camState.around_target_rot_y += 1;
            skyBox->setYRot(skyBox->getYRot() + 1);
        }
    } else if (x > prevX) {
        if (rightButtonDown) {
            camState.around_target_rot_y -= 1;
            skyBox->setYRot(skyBox->getYRot() - 1);
        } 
    }
    
    if (y < prevY) {  
        if (rightButtonDown) {
            if (camState.around_target_rot_x < 90) {
                camState.around_target_rot_x += 1;
                skyBox->setXRot(skyBox->getXRot() + 1);
            }
        } else if (middleButtonDown) {
            if (camState.pos_z > 3) {
                camState.pos_z -= 0.1;
            }
        }           
    } else if (y > prevY) {
        if (rightButtonDown) {
            if (camState.around_target_rot_x > 0) {
                camState.around_target_rot_x -= 1;
                skyBox->setXRot(skyBox->getXRot() - 1);
            }
        } else if (middleButtonDown) {
            if (camState.pos_z < 10) {
                camState.pos_z += 0.1;
            }
        }  
    }  
    

}

void MouseCallback(int button, int state, int x, int y) {
    //check for button presses
    if (button == GLUT_LEFT_BUTTON) {
        if (state == GLUT_UP) {
            leftButtonDown = false;
        } else {
            leftButtonDown = true;
        }
    } else if (button == GLUT_RIGHT_BUTTON) {
        if (state == GLUT_UP) {
            rightButtonDown = false;
        } else  {
            rightButtonDown = true;
        }
    } else if (button == GLUT_MIDDLE_BUTTON) {
        if (state == GLUT_UP) {
            middleButtonDown = false;
        } else  {
            middleButtonDown = true;
        }
    }
    
    pressX = x;
    pressY = y;
    
    if (leftButtonDown && y >= 400 && wizard->isAlive()) {
        hud->handleMouse(x,500-y);
    }
}

int processPick (GLint hits, GLuint buffer[])
{
   GLuint closestNumNames = 0;
   GLuint closestStartOfNames = 0;
   GLuint buffCounter = 0;
   GLuint closest = 0xFFFFFFFF;

   for (int i = 0; i < hits; i++) {
       GLuint numNames = buffer[buffCounter++];
       if (buffer[buffCounter] < closest) {
            closestNumNames = numNames;
            closest = buffer[buffCounter];
            closestStartOfNames = buffCounter + 2;
       }
       buffCounter += numNames + 2;
   }

   if (closestNumNames > 1 || closestNumNames == 0) return -1;
   return buffer[closestStartOfNames];
}

int pick(int x, int y) {
    GLint hits;
    GLuint hitList[256];
    GLint viewport[4];
    
    glGetIntegerv (GL_VIEWPORT, viewport); 
    glSelectBuffer(256, hitList);
    glRenderMode(GL_SELECT);  
   
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluPickMatrix((GLdouble) x, (GLdouble) viewport[3] - y, 5.0,5.0, viewport);
    gluPerspective( 45, 1.0, 0.1, 200.0 );  
    
    glMatrixMode(GL_MODELVIEW);   
    glInitNames();
    glLoadIdentity();  
    
    
    cam.position();
    level->drawForPick();
    
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    glFlush();
    
    hits = glRenderMode(GL_RENDER);
    return processPick(hits, hitList);
}

void DisplayCallback()
{
    glLoadIdentity();
    // clear the color, depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //Set up Perspective Projection
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 45, 1.0, 0.1, 200.0 );  

    glMatrixMode( GL_MODELVIEW );    
    glLoadIdentity();
    
    glColor3f(1,1,1);
    skyBox->draw(); //draw the skybox
    
    glPushMatrix();
        cam.position();
        glLightfv(GL_LIGHT0, GL_POSITION, light0Pos);
        GLfloat pos[4];
        pos[0] = wizard->getPositionX();
        pos[1] = wizard->getPositionY() + 2;
        pos[2] = wizard->getPositionZ();
        pos[3] = 1;
        glLightfv(GL_LIGHT1, GL_POSITION, pos);
    	glEnable(GL_CULL_FACE);
        glColor3f(1,1,1);
        level->draw();
        wizard->draw();        
        glDisable(GL_CULL_FACE);
    glPopMatrix();

    glPushMatrix();
        
        glRotatef(skyBox->getXRot(),1,0,0);
        glRotatef(skyBox->getYRot(),0,1,0);
        rain->draw();
    glPopMatrix();
  
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glOrtho(0,500,0,500,0,1);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glDisable( GL_LIGHTING );
    hud->draw();
    glEnable( GL_LIGHTING );

     // draw the buffer to the screen
    glutSwapBuffers();
    
}

void init() {
    srand(time(NULL));
    
    //Register Callbacks
    glutDisplayFunc( DisplayCallback );
    glutKeyboardFunc( KeyboardCallback );
    glutKeyboardUpFunc( KeyboardUpCallback );
    glutIdleFunc( IdleCallback);
    glutMotionFunc( MouseMotionCallback);
    glutMouseFunc(MouseCallback);
    glutSpecialFunc(SpecialKeyCallback);
    glutSpecialUpFunc(SpecialKeyUpCallback);
    
    glShadeModel(GL_SMOOTH); 
    glEnable(GL_NORMALIZE);
    glEnable(GL_TEXTURE_2D);
    glEnable( GL_LIGHTING );
    glEnable( GL_COLOR_MATERIAL );
    glEnable( GL_DEPTH_TEST );
    glDepthFunc(GL_LEQUAL);
    glClearColor(0,0,0,0);
    glLineWidth(2);
    glCullFace(GL_BACK);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    //set up light
    glLightfv( GL_LIGHT0, GL_AMBIENT, ambient0);
    glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuse0);
    glLightfv( GL_LIGHT0, GL_SPECULAR, specular);
    glEnable( GL_LIGHT0);
    
    glLightfv( GL_LIGHT1, GL_AMBIENT, ambient1);
    glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse1);
    glLightfv( GL_LIGHT1, GL_SPECULAR, specular);
    glEnable( GL_LIGHT1);
    
    //Set up viewport   
    glViewport(0,0,500,500);
    
    detector = new Detector();
    
    //load game models
    wizard = new Wizard(detector);
    hud = new Hud(wizard);
    
    level = new Level(detector,wizard);
    
    skyBox = new Skybox("Textures/front.tga", "Textures/back.tga", "Textures/left.tga",
              "Textures/right.tga", "Textures/up.tga", "Textures/down.tga");
    
    skyBox->setXRot(20);
    skyBox->setYRot(135);
    skyBox->setZRot(0);   
    
    rain  = new RainParticleSystem(500, 0.01, 1, 5, 0.5,
                    1, 1, 1,
                    1, -1, -1, 1, 1, -1); 
     
    camState.target_x = wizard->getPositionX();
    camState.target_y = wizard->getPositionY();
    camState.target_z = wizard->getPositionZ();
    camState.pos_z = 5;
    camState.pos_y = 1;
    camState.around_target_rot_y = 135;
    camState.around_target_rot_x = 20;
      
}



int main(int argc, char **argv)
{
    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_DEPTH | GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowSize( 500, 500 );
    glutCreateWindow( "Wizzow" );   
      
    init();
  
    glutMainLoop();  
    return 0;
}
